{-
Repraesentation eines flexiblen Arrays
-}

module FlexArray(Flex, empty, isEmpty, access, update, hiext, hirem, loext, lorem)
where
  data Flex a = Null | Leaf a | Fork Int (Flex a) (Flex a)
  
  --leeres Array
  empty :: Flex a
  empty = Null
  
  --Test, ob das Array leer ist
  isEmpty :: Flex a -> Bool
  isEmpty Null           = True
  isEmpty (Leaf x)       = False
  isEmpty (Fork n xt yt) = False
  
  --aktuelle Groesse des Arrays
  size :: Flex a -> Int
  size Null           = 0
  size (Leaf x)       = 1
  size (Fork n xt yt) = n
  
  --Test, ob der Baum ein Blatt-Baum ist
  isLeafTree :: Flex a -> Bool
  isLeafTree xt = isEmpty xt || isLeafy xt
    where  
    isLeafy :: Flex a -> Bool
    isLeafy Null           = False
    isLeafy (Leaf x)       = True
    isLeafy (Fork n xt yt) = isLeafy xt && isLeafy yt
  
  --Hoehe des Baums
  height :: Flex a -> Int
  height Null           = 0
  height (Leaf x)       = 0
  height (Fork n xt yt) = 1 + (height xt `max` height yt)
  
  --Zugriff auf ein bestimmtes Element
  access :: Flex a -> Int -> a
  access (Leaf x) 0       = x
  access (Fork n xt yt) k = if k < m then access xt k
                            else access yt (k - m)
                            where
                              m = size xt
  
  --Aktualisieren eines Elements
  update :: Flex a -> Int -> a -> Flex a
  update (Leaf y) 0 val = Leaf val
  update (Fork n xt yt) pos val
    |(pos < m)   = Fork n (update xt pos val) yt
    |otherwise   = Fork n xt (update yt (pos - m) val)
      where m    = size xt
  
  --Hinzufuegen eines Elements am Schwanz
  hiext :: a -> Flex a -> Flex a
  hiext x Null           = Leaf x
  hiext x (Leaf y)       = Fork 2 (Leaf y) (Leaf x)
  hiext x (Fork n xt yt) = Fork (n + 1) (Fork n xt yt) (Leaf x)
  
  --verbesserte hiext (einseitige Flexibilitaet)
  hiext' :: a -> Flex a -> Flex a
  hiext' x Null     = Leaf x
  hiext' x (Leaf y) = Fork 2 (Leaf y) (Leaf x)
  hiext' x (Fork n xt yt)
    |(size xt == size yt) = Fork (n + 1) (Fork n xt yt) (Leaf x)
    |otherwise            = Fork (n + 1) xt (hiext x yt)

  --noch besseres hiext (volle Flexibilitaet)
  hiext'' :: a -> Flex a -> Flex a
  hiext'' x Null           = Leaf x
  hiext'' x (Leaf y)       = Fork 2 (Leaf y) (Leaf x)
  hiext'' x (Fork n xt yt) = Fork (n + 1) xt (hi x yt)
    where
    hi :: a -> Flex a -> Flex a
    hi x (Leaf y) = Fork 2 (Leaf y) (Leaf x)
    hi x (Fork n xt yt)
      |(size xt == size yt) = Fork (n + 1) (Fork n xt yt) (Leaf x)
      |otherwise            = Fork (n + 1) xt (hi x yt)
  
  --Loeschen des letzten Elements
  hirem :: Flex a -> Flex a
  hirem (Leaf x) = Null
  hirem (Fork (n + 1) xt yt) = fork n xt (hirem yt)
  
  --Anfuegen an den "Kopf" des Arrays
  loext :: a -> Flex a -> Flex a      
  loext x Null           = Leaf x
  loext x (Leaf y)       = Fork 2 (Leaf x) (Leaf y)
  loext x (Fork n xt yt) = Fork (n + 1) (lo x xt) yt
    where
    lo :: a -> Flex a -> Flex a
    lo x (Leaf y) = Fork 2 (Leaf x) (Leaf y)
    lo x (Fork n xt yt)
      |(size xt == size yt) = Fork (n + 1) (Leaf x) (Fork n xt yt)
      |otherwise            = Fork (n + 1) (lo x xt) yt
    
  --Loeschen des ersten Elements
  lorem :: Flex a -> Flex a
  lorem (Leaf x)             = Null
  lorem (Fork (n + 1) xt yt) = fork n (lorem xt) yt
  
  --Erzeugen eines Knotens  
  fork :: Int -> Flex a -> Flex a -> Flex a
  fork n xt yt
    |isEmpty xt = yt
    |isEmpty yt = xt
    |otherwise  = Fork n xt yt
   
  
  reprn :: Flex a -> [a]
  reprn Null           = []
  reprn (Leaf x)       = [x]
  reprn (Fork n xt yt) = reprn xt ++ reprn yt
  
  abstr :: [a] -> Flex a
  abstr = foldr hiext'' empty . reverse
        
      
  
  instance Show a => Show (Flex a)
    where
    show Null        =  "-"
    show (Leaf x) = show x ++ " "
    show (Fork h xt zt) =  "(" ++ show h ++ "/"
                    ++ "l:" ++ show xt                    
                    ++ "r:" ++ show zt ++ ")" 
                                                {-
  instance Show a => Show (Flex a)
    where
    show Null = "{}"
    show tree = show $ reprn tree             -}