{-
Repraesentation einer Menge in Form eines balancierten Baumes
-}

module SetByBTree(BTSet, empty, isEmpty, member, insert, delete)
where
  data BTSet a = Null | Fork Int (BTSet a) a (BTSet a)
  
  --Erstellen eines Knotens
  fork :: BTSet a -> a -> BTSet a -> BTSet a
  fork xt y zt = Fork h xt y zt
                 where h = 1 + (max (ht xt) (ht zt))
  
  --Erzeugen eines quasi-perfekten gren-erweiterten Baumes
  spoon :: BTSet a -> a -> BTSet a -> BTSet a
  spoon xt y zt
    |(hz + 1 < hx) && (bias xt < 0) = rotr(fork (rotl xt) y zt)
    |(hz + 1 < hx) = rotr(fork xt y zt)
    |(hx + 1 < hz) && (0 < bias zt) = rotl(fork xt y (rotr zt))
    |(hx + 1 < hz) = rotl(fork xt y zt)
    |otherwise = fork xt y zt
    where
      --Hoehe xt
      hx = ht xt
      --Hoehe zt
      hz = ht zt
      --Differenz (Hoehe xt) und (Hoehe zt)
      bias :: BTSet a -> Int
      bias (Fork h xt y zt) = ht xt - ht zt
      --Baum nach rechts rotieren
      rotr :: BTSet a -> BTSet a
      rotr (Fork m (Fork n ut v wt) y zt) = fork ut v (fork wt y zt)
      --Baum nach links rotieren
      rotl :: BTSet a -> BTSet a
      rotl (Fork m ut v (Fork n rt s tt)) = fork (fork ut v rt) s tt  
  
  --Hoehe eines Baumes
  ht :: BTSet a -> Int
  ht Null = 0
  ht (Fork h xt y zt) = h

  --leere Menge    
  empty :: BTSet a
  empty = Null
  
  --Test auf leere Menge
  isEmpty :: BTSet a -> Bool
  isEmpty Null             = True
  isEmpty (Fork h xt y zt) = False  
  
  --Test, ob Element in Menge enthalten ist
  member :: (Ord a) => BTSet a -> a -> Bool
  member Null x = False
  member (Fork h xt y zt) x
    |(x < y)  = member xt x
    |(x == y) = True
    |(x > y)  = member zt x
  
  --Einfuegen eines Elements in die Menge
  insert :: (Ord a) => a -> BTSet a -> BTSet a
  insert x Null = fork Null x Null
  insert x (Fork h xt y zt)
    |(x < y) = spoon (insert x xt) y zt
    |(x == y) = Fork h xt y zt
    |(x > y) = spoon xt y (insert x zt)
  
  --loeschen eines Elements aus der Menge
  delete :: (Ord a) => a -> BTSet a -> BTSet a
  delete x Null = Null
  delete x (Fork h xt y zt)
    |(x < y)  = spoon (delete x xt) y zt
    |(x == y) = join xt zt
    |(x > y)  = spoon xt y (delete x zt)
  
  
  --Hilfsfunktionen
  join :: BTSet a -> BTSet a -> BTSet a
  join xt yt = if isEmpty yt then xt
               else spoon xt y zt
               where
                 (y,zt) = splitTree yt
    
  splitTree :: BTSet a -> (a, BTSet a)
  splitTree (Fork h xt y zt) = if isEmpty xt then (y, zt) 
                               else (u, spoon vt y zt)
                               where
                                 (u, vt) = splitTree xt   
  
  
  abstr :: (Ord a) => [a] -> BTSet a
  abstr = foldr insert empty . reverse
  
  
  reprn :: BTSet a -> [a]
  reprn Null = []
  reprn (Fork h xt y zt) = reprn xt ++ [y] ++ reprn zt
  
  
  instance Show a => Show (BTSet a)
    where
    show Null        =  "-"
    show (Fork h xt y zt) =  "(Fork "
                    ++ show xt ++ " " 
                    ++ show y ++ " "
                    ++ show zt ++ ")"
  
  {-
  instance Show a => Show (BTSet a)
    where
    show Null = "{}"
    show tree = '{':show' tree ++ "}"
      where
      show' Null = ""
      show' (Fork h xt y zt) = show' xt ++ show y ++ " " ++ show' zt   -}