{-
  Repraesenation eines Sets unter Verwendung von Listen
-}
                      
module SetByList(LSet, empty, isEmpty, insert, member, delete, union, minus, meet)
where
  newtype LSet a = MkLSet ([a])

  --Dictionary-Funktionen
  
  --leere Menge
  empty :: LSet a
  empty = MkLSet []
  
  --Test auf leere Menge
  isEmpty :: LSet a -> Bool
  isEmpty (MkLSet x) = null x
    
  --Test, ob Element in Menge ist
  member :: (Eq a) => LSet a -> a -> Bool
  member xs x = some (==x) xs  

  --verbesserte Member-Funktion, arbeitet auf Basis einer sortierten Liste
  member' :: (Ord a) => LSet a -> a -> Bool
  member' (MkLSet xs) x = if null ys then False
                          else (x == head ys)
		                      where
                            ys = dropWhile (<x) $ xs
  
  --loescht ein Element aus der Menge
  delete :: (Eq a) => a -> LSet a -> LSet a
  delete x (MkLSet xs) = MkLSet(filter (/=x) xs)

  --fuegt ein Element der Menge hinzu
  insert :: a -> LSet a -> LSet a
  insert x (MkLSet xs) = MkLSet(x:xs)
  
  --verbesserte insert-Funktion fuer Menge ohne Duplikate
  insert' :: (Eq a) => a -> LSet a -> LSet a
  insert' x (MkLSet xs) = MkLSet(x:(filter (/=x) xs))
  
  --insert-Funktion ohne Duplikate, sortiert
  insert'' :: (Ord a) => a -> LSet a -> LSet a
  insert'' x (MkLSet xs) = MkLSet(ys ++ [x] ++ (filter (/=x) zs))
                             where
                             (ys, zs) = span (<x) xs  


  --weitere Funktionen
  --Vereinigung von zwei Mengen
  union :: LSet a -> LSet a -> LSet a
  union (MkLSet xs) (MkLSet ys) = MkLSet(xs ++ ys)

  --verbesserte union-Funktion fuer Menge ohne Duplikate
  union' :: (Eq a) => LSet a -> LSet a -> LSet a
  union' (MkLSet xs) (MkLSet ys) = MkLSet(xs ++ filter (not . member (MkLSet xs)) ys)

  --noch bessere union-Funktion
  union'' :: (Ord a) => LSet a -> LSet a -> LSet a
  union'' (MkLSet []) ys = ys
  union'' xs (MkLSet []) = xs
  union'' (MkLSet(x:xs)) (MkLSet(y:ys))
    |(x < y)  = MkLSet (x:reprn (union'' (MkLSet xs) (MkLSet(y:ys))))
    |(x == y) = MkLSet (x:reprn (union'' (MkLSet xs) (MkLSet ys)))
    |(x > y)  = MkLSet (y:reprn (union'' (MkLSet (x:xs)) (MkLSet ys)))
  
  --Differenz von zwei Mengen
  minus :: (Eq a) => LSet a -> LSet a -> LSet a
  minus (MkLSet xs) (MkLSet ys) = MkLSet(filter (not . member (MkLSet ys)) xs)
  
  minus' :: (Ord a) => LSet a -> LSet a -> LSet a
  minus' xs (MkLSet []) = xs
  minus' (MkLSet []) ys = MkLSet [] 
  minus' (MkLSet (x:xs)) (MkLSet (y:ys))        
    |(x < y) = MkLSet (x:reprn(minus' (MkLSet xs) (MkLSet (y:ys))))
    |(x == y) = minus' (MkLSet xs) (MkLSet ys)
    |(x > y) =  minus' (MkLSet (x:xs)) (MkLSet ys)
  
  
  --Schnittmenge
  meet :: (Ord a) => LSet a -> LSet a -> LSet a
  meet xs (MkLSet [])    = empty
  meet xs (MkLSet(y:ys)) = if member xs y then insert y (meet xs (MkLSet ys))
                           else meet xs (MkLSet ys)
   
  meet' :: (Ord a) => LSet a -> LSet a -> LSet a
  meet' (MkLSet []) ys = MkLSet [] 
  meet' xs (MkLSet []) = MkLSet []
  meet' (MkLSet (x:xs)) (MkLSet (y:ys))
    |(x < y)  = meet' (MkLSet xs) (MkLSet(y:ys))
    |(x == y) = MkLSet (x:reprn (meet' (MkLSet xs) (MkLSet ys)))
    |(x > y)  = meet' (MkLSet (x:xs)) (MkLSet ys)


  --Hilfsfunktionen
  some :: (a -> Bool) -> LSet a -> Bool
  some p xs = or . map p $ reprn xs
  
  abstr :: (Ord a) => [a] -> LSet a
  abstr = foldr insert'' empty
  

  reprn :: LSet a -> [a]
  reprn (MkLSet []) = []
  reprn (MkLSet(x:xs))
    |null xs   = [x]
    |otherwise = [x] ++ reprn (MkLSet xs)
  
  
  --Ausgabe eines Sets in der ueblichen Form {[Werte]}
  instance Show a => Show (LSet a)
    where
    show (MkLSet xs)
      |null xs = "{}"
      |otherwise = '{':show' (MkLSet xs) ++ "}"
      where
        show' (MkLSet([])) = "" 
        show' (MkLSet(x:xs)) = show x ++ " " ++ show' (MkLSet xs)
            